#%%
from obspy.clients.fdsn import Client
from obspy.core import UTCDateTime, read, Stream
from obspy import read_events

from obspy.io.sac.sactrace import SACTrace
from obspy import Catalog

import numpy as np
import obspy as op

import matplotlib.path as mpltPath

from numba import jit


import multiprocessing
from functools import partial


import csv
import sys
import os
import re
import json
from pathlib import Path
import logging
import argparse    

log = logging.getLogger(__name__)

#%%
def get_mseedid(starttime):
    yearOUT = starttime.year
    jdayOUT = starttime.julday
    hourOUT = starttime.hour
    minOUT = starttime.minute
    secOUT = starttime.second

    yearOUT2 = (str)(yearOUT)
        
    jdayOUT2 = (str)(jdayOUT)
    if jdayOUT < 100:
        jdayOUT2 = "0"+(str)(jdayOUT)
            

    if jdayOUT < 10:
        jdayOUT2 = "00"+(str)(jdayOUT)
            

    hourOUT2 = (str)(hourOUT)
    if hourOUT < 10:
        hourOUT2 = "0"+(str)(hourOUT)
            

    minOUT2 = (str)(minOUT)
    if minOUT < 10:
        minOUT2 = "0"+(str)(minOUT)


    secOUT2 = (str)(secOUT)
    if secOUT < 10:
        secOUT2 = "0"+(str)(secOUT)
        
    log.debug("yearOUT2 = {0} jdayOUT2 = {1} hourOUT2 = {2} minOUT2 = {3} secOUT2 = {4}".format(yearOUT2, jdayOUT2, hourOUT2, minOUT2, secOUT2))
    mseedid = yearOUT2+"."+jdayOUT2+"."+hourOUT2+""+minOUT2+""+secOUT2
    return mseedid


def get_evmseedid(origin_time):

    evyearOUT = origin_time.year
    evjdayOUT = origin_time.julday
    evhourOUT = origin_time.hour
    evminOUT = origin_time.minute
    evsecOUT = origin_time.second

    # need for file name
    evyearOUT2 = (str)(evyearOUT)
    evjdayOUT2 = (str)(evjdayOUT)
    if evjdayOUT < 100:
        evjdayOUT2 = "0"+(str)(evjdayOUT)
            

    if evjdayOUT < 10:
        evjdayOUT2 = "00"+(str)(evjdayOUT)
        

    evhourOUT2 = (str)(evhourOUT)
    if evhourOUT < 10:
        evhourOUT2 = "0"+(str)(evhourOUT)
            

    evminOUT2 = (str)(evminOUT)
    if evminOUT < 10:
        evminOUT2 = "0"+(str)(evminOUT)


    evsecOUT2 = (str)(evsecOUT)
    if evsecOUT < 10:
        evsecOUT2 = "0"+(str)(evsecOUT)
            
    log.debug("evyearOUT2 = {0} evjdayOUT2 = {1} evhourOUT2 = {2} evminOUT2 = {3} evsecOUT2 = {4}".format(evyearOUT2, evjdayOUT2, evhourOUT2, evminOUT2, evsecOUT2))
        
    evmseedid = evyearOUT2+"."+evjdayOUT2+"."+evhourOUT2+""+evminOUT2+""+evsecOUT2

    # name for output directory
    # pwd_dir = os.getcwd() 

    return evmseedid

def get_eq_info(catalog):

    # event info. origin time, location, magnitude
    event = catalog[0]
    origin = event.origins[0]
    origin_time = origin.time
    evlat = origin.latitude
    evlon = origin.longitude
    evdp_km = origin.depth / 1000

    #evmag = event.magnitudes[0].mag
    try:
        evmag = event.magnitudes[0].mag 
    except:
        eventid = event.origins[0]['extra']['eventid']['value']
        log.warning("eventid {0} has no mag info we use mag=-9".format(eventid))
        evmag = -9 # there is a case that we have mag before but now (2023), NCSS will not have mag info.

    return origin_time, evlat, evlon, evdp_km, evmag
        

def download_eq_info(eventid, eq_client_name, eq_client_dict):

    eq_client = Client(eq_client_name, timeout=1200)
    eq_client_para = eq_client_dict.get(eq_client_name)

    eq_found = 0 
    catalog_name = None
    catalog = None

    try:
        relocation_catalog = eq_client_para['relocation_catalog']
    except:
        log.debug("no relocation_catalog eq_client_name = {0}".format(eq_client_name))
        relocation_catalog = False

    if relocation_catalog:
        try:
            log.debug("relocation catalog. eventid = {0}".format(eventid))
            catalog = eq_client.get_events(eventid=eventid, catalog=relocation_catalog)
            eq_found = 2 

        except:
            log.warning("no relocated solution use default catalog. eventid = {0} eq_client_para['default_catalog'] {1}".format(eventid, eq_client_para['default_catalog'] ))
            try:
                catalog = eq_client.get_events(eventid=eventid)
                eq_found = 1 

            except:     
                log.warning("no catalog eventid = {0}".format(eventid))
    else:
        try:
            catalog = eq_client.get_events(eventid=eventid)
            eq_found = 1 
        except:     
            log.warning("no catalog eventid = {0}".format(eventid))

    
    if eq_found >= 1:
        log.debug("event info eventid = {0} found ".format(eventid))
        if eq_found == 1:
            try:
                catalog_name = eq_client_para['default_catalog'] 
            except:
                catalog_name = "default"
        if eq_found == 2:
            catalog_name = relocation_catalog
    

    #evid_name = eq_client_para['evid_name'] 
    #catalog_name = eq_client_para['catalog_name'] 
    #log.debug("evid_name = {0}".format(evid_name))
    log.debug("catalog_name = {0}".format(catalog_name))
    
    return catalog, catalog_name, eq_found



def select_event_joblib(eq_point, event): # Function is compiled to machine code when called the first time
    if eq_point:
        return event      

def select_event_map(task): # Function is compiled to machine code when called the first time
    eq_point = task[0]
    event = task[1]
    if eq_point:
        return event
    else:
        log.debug("removed eq = {0}".format(event))            

@jit(nopython=True) # Set "nopython" mode for best performance, equivalent to @njit
# did not work
# - argument 1: Cannot determine Numba type of <class 'obspy.core.event.catalog.Catalog'>
def select_event(eq_points_inside, catalog_all): # Function is compiled to machine code when called the first time
    catalog_select = Catalog()
    for eq_point, event in zip(eq_points_inside, catalog_all):
        #log.debug("eq_point = {0}".format(eq_point))
        #log.debug("type(eq_point) = {0}".format(type(eq_point)))

        if eq_point:
            #log.debug("append event = {0}".format(event))
            catalog_select.append(event)
    return catalog_select              # Numba likes NumPy broadcasting

def load_jsonc(filepath):
    with open(filepath, 'r') as f:
        text = f.read()
    text_without_comment = re.sub(r'/\*[\s\S]*?\*/|//.*', '', text)
    json_obj = json.loads(text_without_comment)
    return json_obj

def start_parallel_jobs_inner_loop(tasks, do_work, njobs=0):
    if njobs == 0:
        results = [do_work(task) for task in tasks]
    else:
        pool = multiprocessing.Pool(njobs)
        results = pool.map(do_work, tasks)
        pool.close()
        pool.join()
        
    return results

def request_waveforms(fdsn_server_list, network_dict,
                      distdeg_from_eq, eq_client_name, eq_client_dict,
                      f1, f2, f3, f4, output_unit, output_top_dir,
                      pre_tw_sec, tw_sec,
                      task):
    
    #eventid = task
    log.debug("task.origins[0] = {0}".format(task.origins[0]))
    eventid = task.origins[0]['extra']['eventid']['value']
    try:
        datasource = task.origins[0]['extra']['datasource']['value']
    except:
        log.warning("no datasource... eventid {0} use xx now".format(eventid))
        datasource = "xx"

    log.debug("eventid = {0}".format(eventid))
    log.debug("datasource = {0}".format(datasource))

    for fdsn_server in fdsn_server_list:
        log.debug("fdsn_server = {0}".format(fdsn_server))

        #coms_default = "HHZ,EHZ"
        coms_default = "BHZ" # do not hardcoding...
        com_list = coms_default



        network_para = network_dict.get(fdsn_server)
        log.debug("network_para = {0}".format(network_para))         

        net_list = network_para['nets'] 
        log.debug("net_list = {0}".format(net_list))         

        try:
            add_coms = network_para['add_coms']
        except:
            log.debug("no add_coms fdsn_server = {0}".format(fdsn_server))
            add_coms = False
        if add_coms:
            log.debug("add additional coms")
            log.debug("add_coms = {0}".format(add_coms))

            com_list = com_list+","+add_coms

        log.debug("com_list = {0}".format(com_list))
        
        try:    
            download_waveforms(eventid, datasource, distdeg_from_eq, eq_client_name, eq_client_dict,
                               fdsn_server, net_list, com_list,
                               f1, f2, f3, f4, pre_tw_sec, tw_sec,
                               output_unit, output_top_dir)
            
        except Exception as e:
            log.error("cannot preprocess for event {0} at fdsn server {1}".format(eventid, fdsn_server))
            log.error(e)


def download_waveforms(eventid, datasource, distdeg_from_eq, eq_client_name, eq_client_dict,
                       fdsn_server, net_list, com_list,
                       f1, f2, f3, f4, pre_tw_sec, tw_sec,
                       output_unit, output_top_dir):
    try:

        #distdeg_from_eq = 0.4
        #distdeg_from_eq = 0.2
        # search distance from event in degrees
        #distdeg_from_eq = 0.35

        # need for catalog
        #client_scedc = Client("SCEDC")
        # extracting event info. from datacenter catalog
        #client_ncedc = Client("NCEDC",timeout=600)

        #client_ncedc = Client(eq_client_name, timeout=1200)


        try:
            if eq_client_name is "NCEDC":
                eventid_in_db = eventid
            else: 
                eventid_in_db = datasource+""+eventid
            log.debug("eventid_in_db = {0}".format(eventid_in_db))
            catalog, catalog_name, eq_found = download_eq_info(eventid_in_db, eq_client_name, eq_client_dict)
        except:
            log.error("cannot get catalog for eventid {0} from eq_client_name {1}".format(eventid, eq_client_name))

        if eq_found >= 1:
            origin_time, evlat, evlon, evdp_km, evmag = get_eq_info(catalog)
            evmseedid = get_evmseedid(origin_time)
            log.debug("evmseedid = {0}".format(evmseedid))
        
            #sacdir= pwd_dir +"/"+ evmseedid +"_M"+(str)(evmag)+"_"+(str)(eventid_ncedc)+"_fl"+(str)(f2)+"_fh"+str(f3)+"_dist"+(str)(distkm_from_eq)+"km"
            #sacdir = pwd_dir +"/"+ evmseedid +"_M"+(str)(evmag)+"_"+datasource+(str)(eventid)+"_fl"+(str)(f2)+"_fh"+str(f3)+"_dist"+(str)(distdeg_from_eq)+"deg"
            sacdir = output_top_dir+"/"+ evmseedid +"_M"+(str)(evmag)+"_"+datasource+(str)(eventid)+"_fl"+(str)(f2)+"_fh"+str(f3)+"_dist"+(str)(distdeg_from_eq)+"deg"


            log.debug("sacdir = {0}".format(sacdir))

            # create output directory
            if not os.path.exists(sacdir):
                os.makedirs(sacdir)

            # time window from origin time to extract waveform
            #pre_tw_sec = 90 # 90 seconds before the origin time
            #tw_sec = 5*60 # 5 min after the origin time


            starttime = UTCDateTime(origin_time-pre_tw_sec)
            endtime = UTCDateTime(origin_time+tw_sec)

            #starttime2 = UTCDateTime(origin_time-86400)
            #endtime2 = UTCDateTime(origin_time+86400)

            starttime2 = UTCDateTime(origin_time-pre_tw_sec)
            endtime2 = UTCDateTime(origin_time+tw_sec)

            log.debug("starttime  = {0} endtime  = {1}".format(starttime, endtime))
            log.debug("starttime2 = {0} endtime2 = {1}".format(starttime2, endtime2))

            mseedid = get_mseedid(starttime)
            log.debug("mseedid = {0}".format(mseedid))

            # this would be only need for SCEDC. SSAF project
            # change all m type to this. we have error "SoCal ML"
            #catalog_scedc[0].magnitudes[0].method_id = "smi:ci.anss.org/magnitude/CISNml2"

            # write event information in QUAKEML format
            cat_fi = sacdir+"/"+datasource+(str)(eventid)+".eq.xml"
            log.debug("cat_fi = {0}".format(cat_fi))
            catalog.write(cat_fi, format="QUAKEML")  

            log.debug("invOUT fdsn_server = {0}".format(fdsn_server))
            log.debug("net_listOUT = {0} com_listOUT = {1}".format(net_list, com_list))

            # extracting stations within min and max radius, also get instrument response
            #invOUT = clientOUT.get_stations(network=net_listOUT, station="*", channel=com_listOUT, 
            #                                starttime=starttime, endtime=endtime,
            #                                latitude=latitude, longitude=longitude,
            #                                minradius=minradius, maxradius=maxradius,
            #                                level="response")

            # if we have same station code in different networks, this will provide both. it is a bit problem but we ignore this moment. 06.12.23
            # use time2 +/- 1 days to fetch multiple time epochs if we have. there is a bug for FDSN code cannot handle the epoch change 05.07.19

            latitude = evlat
            longitude = evlon
            minradius = 0
            maxradius = distdeg_from_eq

            wf_client = Client(fdsn_server, timeout=1200)

            inv_found = 0
            try:
                invOUT = wf_client.get_stations(network=net_list, station="*", channel=com_list, 
                                                starttime=starttime2, endtime=endtime2,
                                                latitude=latitude, longitude=longitude,
                                                minradius=minradius, maxradius=maxradius,
                                                level="response")
                inv_found = 1
            except:
                log.warning("no inv found eventid {0} fdsn_server {1}".format(eventid, fdsn_server))

            channel_found = 0
            try:
                channelOUT = wf_client.get_stations(network=net_list, station="*", channel=com_list, 
                                                starttime=starttime2, endtime=endtime2,
                                                latitude=latitude, longitude=longitude,
                                                minradius=minradius, maxradius=maxradius,
                                                level="channel")
                channel_found = 1

            except:
                log.warning("no channel found eventid {0} fdsn_server {1}".format(eventid, fdsn_server))



            if inv_found and channel_found:
                # get station list
                ch_listOUT = ""
                j2=0
                sncl_list = []
                for i in range(len(channelOUT)):        
                    netOUT = channelOUT[i].code
                    #print (inv_ncedc[i].code)
                    for ii in range(len(channelOUT[i])):
                        #print(ii)
                        #print(inv[i][ii].code)    
                        staOUT = channelOUT[i][ii].code
                        for iii in range(len(channelOUT[i][ii])):
                            chaOUT = channelOUT[i][ii][iii].code 
                            locOUT = channelOUT[i][ii][iii].location_code
                            snclOUT = netOUT+"."+staOUT+"."+locOUT+"."+chaOUT
                            #print("# j2 = ",j2," snclOUT = ",snclOUT)
                            sncl_list.append(snclOUT)
                            j2+=1
                        

                # write station metadata in STATIONXML format
                #inv_fiOUT = sacdir + "/" + (str)(eventid_ncedc)+".sta."+(str)(distdeg_from_eq)+"km."+server_nameOUT+".xml"
                inv_fiOUT = sacdir+"/"+datasource+(str)(eventid)+".sta."+(str)(distdeg_from_eq)+"deg."+fdsn_server+".xml"
                log.debug("inv_fiOUT station xml = {0}".format(inv_fiOUT))
                invOUT.write(inv_fiOUT, format='STATIONXML')  

                inv_fiOUT = sacdir+"/"+datasource+(str)(eventid)+".sta."+(str)(distdeg_from_eq)+"deg."+fdsn_server+".sacpz"
                log.debug("inv_fiOUT sacpz = {0}".format(inv_fiOUT))
                invOUT.write(inv_fiOUT, format='SACPZ')  

                # get station list
                sta_listOUT = ""
                j=0
                for i in range(len(invOUT)):
                    #print (inv_ncedc[i].code)
                    for ii in range(len(invOUT[i])):
                        #print(ii)
                        #print(inv[i][ii].code)
                        if j==0:
                                sta_listOUT = invOUT[i][ii].code
                        else:
                                sta_listOUT = sta_listOUT+","+invOUT[i][ii].code
                        j+=1
                    


                log.debug("sta_listOUT = {0}".format(sta_listOUT))

                # get seismic waveforms nased on station & network lists
                wf_found = 0
                try:
                    stOUT = wf_client.get_waveforms(network=net_list, station=sta_listOUT, channel=com_list, location="*",
                                                    starttime=starttime, endtime=endtime)
                    wf_found = 1
                except:
                    log.warning("no wf found eventid {0} fdsn_server {1}".format(eventid, fdsn_server))


                if wf_found:
                    stOUT2 = stOUT.copy()

                    #st2 = stOUT2.select(network="NC")

                    log.debug(stOUT2.__str__(extended=True))

                    for trace in stOUT2:
                        nptsOUT = trace.stats.npts
                        deltaOUT = trace.stats.delta
                        sampHZOUT = 1.0/deltaOUT
                        #print("sampHZOUT = ", sampHZOUT, "nptsOUT = ", nptsOUT)    
                        if nptsOUT < sampHZOUT:
                            stOUT2.remove(trace)



                    log.debug(stOUT2.__str__(extended=True))

                    stOUT22 = stOUT2.copy()
                    stOUT22.clear()  
                    for i in range(len(sncl_list)):
                        log.debug("sncl_list[i] = {0}".format(sncl_list[i]))
                        stOUT22 += stOUT2.select(id=sncl_list[i])


                    stOUT2.clear() 
                    stOUT2 = stOUT22.copy()
                    log.debug(stOUT2.__str__(extended=True))

                    # remove instrument response
                    log.debug("remove_response fdsn_server = {0}".format(fdsn_server))

                    # band pass when we remove instrument response
                    # f1<f2<f3<f4. effective band is between f2 and f3
                    #f1 = 0.1 # 0.1 Hz
                    #f2 = 0.5  # 0.5 Hz
                    #f3 = 40 # 50 Hz
                    #f4 = 50 # 100 Hz


                    stOUT2.remove_response(inventory=invOUT, 
                                    pre_filt=(f1, f2, f3, f4),
                                    output=output_unit, water_level=None)

                    #st2 = stOUT2.select(network="NC",station="N*", location="10")
                    #print(st2.__str__(extended=True))

                    #print("# remove_response server_nameOUT = ",server_nameOUT)
                    #st2.remove_response(inventory=invOUT, 
                    #                   pre_filt=(f1, f2, f3, f4),
                    #                   output="VEL")



                    # save waveforms in SAC format
                    for i in range(len(stOUT2)):
                        #print (st_ncedc2[i])
                        sac = SACTrace.from_obspy_trace(stOUT2[i])
                        sac.lcalda = True
                        
                        netOUT = stOUT2[i].stats.network
                        comOUT = stOUT2[i].stats.channel
                        staOUT = stOUT2[i].stats.station
                        locOUT = stOUT2[i].stats.location

                        # st_starttime is different from each sacfile, but mseedid is used origin_time - pre_tw_sec
                        #st_starttime = st_ncedc2[i].stats.starttime
                        #print("# st_starttime = ",st_starttime)
                            
                        
                        stid = netOUT+"."+staOUT+"."+locOUT+"."+comOUT
                        log.debug("staOUT = {0} netOUT = {1} comOUT = {2} locOUT = {3} stid = {4}".format(staOUT, netOUT, comOUT, locOUT, stid))
                        #print("# stid = "+stid)
                        sta_coordinate = invOUT.get_coordinates(stid, starttime)
                        sac.stla = sta_coordinate['latitude']
                        sac.stlo = sta_coordinate['longitude']
                        sac.stel = sta_coordinate['elevation']
                        sac.evla = evlat
                        sac.evlo = evlon
                        sac.evdp = evdp_km
                        
                        sac.o = origin_time - stOUT2[i].stats.starttime

                        #sac.kuser0 = "ncss"
                        sac.kuser0 = catalog_name
                        sac.kevnm = (str)(eventid)
                        sac.kuser1 = datasource

                        sac.mag = evmag
                        
                        outsacfi = sacdir +"/" + staOUT+"."+netOUT+"."+comOUT+"."+locOUT+".D."+mseedid+"."+datasource+(str)(eventid)+".sac"
                        #print("# outsacfi = "+outsacfi)
                        #st_ncedc[i].write(outsacfi, format="SAC")  
                        sac.write(outsacfi)  
        else:
            log.error("cannot find eventid {0} from eq_client_name {1} eq_found {2}".format(eventid, eq_client_name, eq_found))

    except Exception as e:
        log.error("cannot preprocess for event {0} {1} fdsn_server {2} ".format(datasource, eventid, fdsn_server))
        log.error(e)

def query_waveforms(eq_client_name, fdsn_server_list, network_dict, catalog_file,
                    catalog_create, target_region, eq_polygon_dict, eq_client_dict,
                    distdeg_from_eq,
                    pre_tw_sec, tw_sec, f1, f2, f3, f4,
                    output_unit, output_top_dir, plotOPT, njobs, task_limit):
    
    log.info("eq_client_name = {0}".format(eq_client_name))
    log.info("fdsn_server_list = {0}".format(fdsn_server_list))
    log.info("network_dict = {0}".format(network_dict))
    log.info("catalog_file = {0}".format(catalog_file))

    log.info("catalog_create = {0}".format(catalog_create))
    log.info("target_region = {0}".format(target_region))
    log.info("eq_polygon_dict = {0}".format(eq_polygon_dict))

    log.info("distdeg_from_eq = {0}".format(distdeg_from_eq))
    log.info("pre_tw_sec = {0}".format(pre_tw_sec))
    log.info("tw_sec = {0}".format(tw_sec))
    log.info("f1 = {0} f2 = {1} f3 = {2} f4 = {3}".format(f1, f2, f3, f4))
    log.info("output_unit = {0}".format(output_unit))
    log.info("plotOPT = {0}".format(plotOPT))

    log.info("njobs = {0}".format(njobs))
    log.info("task_limit = {0}".format(task_limit))

    if catalog_file is None and catalog_create == 0:
        log.error("no catalog and catalog_create provided")

    if catalog_create:
        log.info("creating eq catalog (eq_id only needed)...")
        
        target_region_para = eq_polygon_dict.get(target_region)
        log.info("target_region_para = {0}".format(target_region_para))
        target_region_para['poly']
        log.debug("target_region_para['poly'] = {0}".format(target_region_para['poly']))
        log.debug("type(target_region_para['poly']) = {0}".format(type(target_region_para['poly'])))
        polyOUT = target_region_para['poly'].replace("[", "").split("],")
        log.debug("polyOUT = {0}".format(polyOUT))
        if polyOUT[-1] == "":
            polyOUT.pop(-1)
        
        log.debug("polyOUT = {0}".format(polyOUT))
        log.debug("type(polyOUT) = {0}".format(type(polyOUT)))
        log.debug("len(polyOUT) = {0}".format(len(polyOUT)))
        #log.info("polyOUT[0] = {0}".format(polyOUT[0]))

        target_polys = np.zeros((len(polyOUT), 2))

        for i, a in enumerate(target_polys):
            target_polys[i, 0], target_polys[i, 1]  = polyOUT[i].split(",")
        
        log.debug("polys = {0}".format(target_polys))
        log.debug("type(target_polys) = {0}".format(type(target_polys)))

        minlon, minlat = np.amin(target_polys, axis=0)
        maxlon, maxlat = np.amax(target_polys, axis=0)

        eq_client = Client(eq_client_name, timeout=1200)

        
        log.info("minlat = {0} maxlat = {1}".format(minlat, maxlat))
        log.info("minlon = {0} maxlon = {1}".format(minlon, maxlon))

        log.info("minmag = {0} maxmag = {1}".format(target_region_para['minmag'], target_region_para['maxmag']))
        log.info("starttime = {0} endtime = {1}".format(target_region_para['starttime'], target_region_para['endtime']))

        #HFRC_NCSS_stY2017_etY2017_poly1_Mmin0.5.out
        catalog_out_fi = target_region_para['catalog_out_dir']+"/"+target_region+"_"+eq_client_name+"_stY"+str(target_region_para['starttime'])+"_etY"+str(target_region_para['endtime'])+"_Mmin"+str(target_region_para['minmag'])+"_Mmax"+str(target_region_para['maxmag'])+".xml"
        log.info("catalog_out_fi = {0}".format(catalog_out_fi))

        catalog_all = eq_client.get_events(starttime=target_region_para['starttime'], endtime=target_region_para['endtime'],
                                           minmagnitude=target_region_para['minmag'], maxmagnitude=target_region_para['maxmag'],
                                           minlatitude=minlat, maxlatitude=maxlat,
                                           minlongitude=minlon, maxlongitude=maxlon)

        log.debug(catalog_all)
        #event = catalog[0]
        #origin = event.origins[0]
        #evla1 = origin.latitude
        #evlo1 = origin.longitude


        #tasks = list(map(lambda x: str((x.resolve()).parents[0]) +"/"+ (x.resolve()).name  , sac_fis))
        eq_points = list(map(lambda x: (x.origins[0].longitude, x.origins[0].latitude) , catalog_all))

        target_path = mpltPath.Path(target_polys)
        eq_points_inside = target_path.contains_points(eq_points)

        ### test
        #eq_points_inside[1] = False    

        #log.debug("eq_points_inside = {0}".format(eq_points_inside))
        #index_select =  np.where( eq_points_inside == True)[0]
        #index =  np.where( eq_points_inside == True)


        #log.debug("index = {0}".format(index_select))
        #log.debug("type(index) = {0}".format(type(index_select)))
        #log.debug("index[0][0] = {0}".format(index[0][0]))
        #log.debug("type(index[0][0]) = {0}".format(type(index[0][0])))

        # use jit
        # did not work
        # - argument 1: Cannot determine Numba type of <class 'obspy.core.event.catalog.Catalog'>
        #catalog_select = select_event(eq_points_inside, catalog_all)

        log.info("len(catalog_all) = {0}".format(len(catalog_all)))
        task_eq = list(zip(eq_points_inside, catalog_all))
        log.debug("task_eq[0] = {0}".format(task_eq[0]))
        log.debug("type(task_eq[0]) = {0}".format(type(task_eq[0])))

        results = start_parallel_jobs_inner_loop(task_eq, select_event_map, njobs)
        log.debug("len(results) = {0}".format(len(results)))
        log.debug("type(results[0]) = {0}".format(type(results[0])))
        log.debug("type(results) = {0}".format(type(results)))

        results_filter = [x for x in results if x is not None]

        catalog_select = Catalog()
        # assuming filterout data less loop here... if not simple_loop is fine
        for event in results_filter:
            catalog_select.append(event)

        log.info("len(catalog_select) = {0}".format(len(catalog_select)))
        log.debug("type(catalog_select[0]) = {0}".format(type(catalog_select[0])))

        log.debug(catalog_select)
        #sys.exit()

        simple_loop = 0
        if simple_loop:
            for eq_point, event in zip(eq_points_inside, catalog_all):
                #log.debug("eq_point = {0}".format(eq_point))
                #log.debug("type(eq_point) = {0}".format(type(eq_point)))

                if eq_point:
                    #log.debug("append event = {0}".format(event))
                    catalog_select.append(event)


        log.debug(catalog_select)

        catalog_select.write(catalog_out_fi, format="QUAKEML")
        tasks = catalog_select

    else:
        # catalog_file
        tasks = read_events(catalog_file, format="QUAKEML")

    log.info("len(tasks) = {0}".format(len(tasks)))

    log.debug("tasks[0] = {0}".format(tasks[0]))
    log.debug("type(tasks[0]) = {0}".format(type(tasks[0])))

    #sys.exit()
    do_work = partial(request_waveforms,
                      fdsn_server_list, network_dict,
                      distdeg_from_eq, eq_client_name, eq_client_dict,
                      f1, f2, f3, f4, output_unit, output_top_dir, 
                      pre_tw_sec, tw_sec)
        
    #task_limit = 10
    if len(tasks) > 0:
        try:
            for i in range(0, len(tasks), task_limit):
                log.info("slice from i {0} to i+task_limit {1} ".format(i, i+task_limit))
                task_subset = (tasks[i: i+task_limit])
                log.debug("task_subset[0] = {0}".format(task_subset[0]))

                results = start_parallel_jobs_inner_loop(task_subset, do_work, njobs)

        except Exception as e:
            log.error("cannot downalod for all events")
            log.error(e)

#%%
def main():
    parser = argparse.ArgumentParser(description='CRE_waveform')   
    parser.add_argument('-c', '--catalog_file', help='catalog file', default=None)
    parser.add_argument('-ecl', '--eq_client_list', help='eq client list', default="CRE_eq_client_list.json")   
    parser.add_argument('-epl', '--eq_polygon_list', help='eq polygon list', default="CRE_eq_polygon_list.json")   
    parser.add_argument('-p', '--parameter_file', help='parameter file', default="CRE_waveform.jsonc")   
    parser.add_argument('-nl', '--network_list', help='network list', default="CRE_network_list.json")   
    parser.add_argument('--debug', action='store_true')

    args = parser.parse_args()    
    
    logging_level = "INFO"
    if args.debug:
        logging_level = "DEBUG"
  
    logging.basicConfig(level=logging_level, format="%(asctime)s\t%(name)s\t%(levelname)s\t%(message)s")
    log.info(log)
    #log.info("args.trace_id = {0}".format(args.trace_id))


    #inp_fi = "CRE_waveform.jsonc"
    inp_fi = args.parameter_file
    log.info("inp_fi = {0}".format(inp_fi))

    inp_fd = open(inp_fi, mode='r')
    
    #inp_para = json.load(inp_fd)
    inp_para = load_jsonc(inp_fi)
    inp_fd.close()
    log.info('inp_para = {0}'.format(inp_para))
    
    network_list_fi = args.network_list
    log.info("network_list_fi = {0}".format(network_list_fi))
    network_dict = {}
    with open(network_list_fi, mode="r") as f:
        network_dict = json.load(f)
    log.info("network_dict = {0}".format(network_dict))


    eq_polygon_list_fi = args.eq_polygon_list
    log.info("eq_polygon_list_fi = {0}".format(eq_polygon_list_fi))
    eq_polygon_dict = {}
    with open(eq_polygon_list_fi, mode="r") as f:
        eq_polygon_dict = json.load(f)
    log.info("eq_polygon_dict = {0}".format(eq_polygon_dict))

    eq_client_list_fi = args.eq_client_list
    log.info("eq_client_list_fi = {0}".format(eq_client_list_fi))
    eq_client_dict = {}
    with open(eq_client_list_fi, mode="r") as f:
        eq_client_dict = json.load(f)
    log.info("eq_client_dict = {0}".format(eq_client_dict))


    fdsn_server_list = inp_para['fdsn_server_list'].split(",")

    #sys.exit()
    query_waveforms(
        inp_para['eq_client_name'], fdsn_server_list, network_dict, args.catalog_file,
        inp_para['catalog_create'], inp_para['target_region'], eq_polygon_dict, eq_client_dict,
        inp_para['distdeg_from_eq'], 
        inp_para['pre_tw_sec'], inp_para['tw_sec'], 
        inp_para['f1'], inp_para['f2'], inp_para['f3'], inp_para['f4'], 
        inp_para['output_unit'], inp_para['output_top_dir'], inp_para['plotOPT'],
        inp_para['njobs'], inp_para['task_limit'])

if __name__ == "__main__":
    main()

# %%
