#%%
from obspy.core import read
from obspy.core import UTCDateTime

import numpy as np
import scipy as sp
import obspy as op
import matplotlib as mpl
import pandas as pd
from scipy import signal
from scipy import stats
import matplotlib.pyplot as plt
from matplotlib import mlab
from scipy import signal

import pyproj as ppj
from numba import jit

import multiprocessing
from functools import partial


import csv
import sys
import os
import re
import json
from pathlib import Path
import logging
import argparse    

log = logging.getLogger(__name__)

 
#%%

@jit(nopython=True) # Set "nopython" mode for best performance, equivalent to @njit
def get_tasks(sac_fis, search_opt): # Function is compiled to machine code when called the first time
    tasks = []
    start_j = 0
    # we do not use master and other files anymore. creating them to seach all available sac. for this seach_opt should be 1. same as master and other files are the same
    for i, master_ev_fi in enumerate(sac_fis):
        if search_opt:
            start_j = i + 1
        for j, others_ev_fi in enumerate(sac_fis[start_j:], start_j):
            #log.debug("i = {0} j = {1}".format(i, j))
            #tasks.append((data_directory+"/"+master_ev_fi.name, data_directory+"/"+others_ev_fi.name))
            #tasks.append((master_ev_fi.resolve(), others_ev_fi.resolve()))
            tasks.append((master_ev_fi, others_ev_fi))

    return tasks              # Numba likes NumPy broadcasting


def load_jsonc(filepath):
    with open(filepath, 'r') as f:
        text = f.read()
    text_without_comment = re.sub(r'/\*[\s\S]*?\*/|//.*', '', text)
    json_obj = json.loads(text_without_comment)
    return json_obj

def corr_norm_stat(data1, data2):
    #data1_norm = (data1 - np.mean(data1)) / (np.std(data1) * len(data1))
    #data2_norm = (data2 - np.mean(data2)) / (np.std(data2))
    #corr_norm = np.correlate(data1_norm, data2_norm, 'full')
    #corr_norm_pos_max = np.max(corr_norm)
    #corr_norm_pos_max_outINT = "{:d}".format(int(corr_norm_pos_max*10000))
    
    cc_spearmanr, p_spearmanr = stats.spearmanr(data1, data2)
    cc_spearmanr_outINT = "{:d}".format(int(cc_spearmanr*10000))

    cc_pearsonr, p_pearsonr = stats.pearsonr(data1, data2)
    cc_pearsonr_outINT = "{:d}".format(int(cc_pearsonr*10000))

        
    #return corr_norm, corr_norm_pos_max_outINT
    return cc_spearmanr_outINT, p_spearmanr, cc_pearsonr_outINT, p_pearsonr


def coh_stat(freq, cyx, fl, fh):
    index =  np.where( (freq >= fl) & (freq <= fh)   )[0]
    #print(index)
    coh_mean = np.mean(cyx[index])
    coh_mean_outINT = "{:d}".format(int(coh_mean*10000))

    return coh_mean, coh_mean_outINT



def fn_info(sac_fi_name):
    sta, net, com, loc, qual, yy, doy, hhmmss, evid, _ = sac_fi_name.split('.')
    mseed_id = yy+"."+doy+"."+hhmmss
    return mseed_id



def sac_header_info(tr):
    evla = tr.stats.sac['evla']
    evlo = tr.stats.sac['evlo']
    evdp = tr.stats.sac['evdp']

    evid = tr.stats.sac['kevnm']
    mag = tr.stats.sac['mag']

    
    stla = tr.stats.sac['stla']
    stlo = tr.stats.sac['stlo']

    
    return evla, evlo, evdp, mag, evid, stla, stlo



def cal_coh(fl, fh, fl1, fh1,
            fl2 ,fh2, fl3, fh3, fl4, fh4, 
            pre_tw_xc, tw_xc_static, max_ev_dist_km,
            plotOPT, 
            nperseg, noverlap,
            diff_fs_eps, 
            tw_xc_dynamic, tw_xc_vs, tw_xc_opt,
            task):
    

    st1_fi, st2_fi = task

    #sacfi_name1 = st1_fi.name.split('/')[-1]
    #sacfi_name2 = st2_fi.name.split('/')[-1]
    sacfi_name1 = st1_fi.split('/')[-1]
    sacfi_name2 = st2_fi.split('/')[-1]

    #log.debug("sacfi_name1 = {0}".format(sacfi_name1))
    #log.debug("type(sacfi_name1) = {0}".format(type(sacfi_name1)))


    try:
        st1 = read(st1_fi)
        st2 = read(st2_fi)
        
        log.debug("st1 = {0}".format(st1))
        log.debug("st2 = {0}".format(st2))

        #mseed_id1 = fn_info(st1_fi)
        #mseed_id2 = fn_info(st2_fi)
        # stf_fi
        mseed_id1 = fn_info(sacfi_name1)
        mseed_id2 = fn_info(sacfi_name2)


        log.debug("st1_fi = %s", st1_fi)
        log.debug("st2_fi = %s", st2_fi)
        log.debug("seed_id1 = %s", mseed_id1)
        log.debug("mseed_id2 = %s", mseed_id2)

        log.debug("st1 = %s", st1)
        log.debug("st2 = %s", st2)


        if plotOPT:
            _plot = st1.plot()
            _plot = st2.plot()
            
            
        tr1=st1[0]
        tr2=st2[0]

        

        try:
            origin_time_diff_tr1_tr2 = (tr1.stats.starttime + tr1.stats.sac['o']) - tr2.stats.starttime + tr2.stats.sac['o']
        except Exception as e:
            # in case origin time is not listed
            origin_time_diff_tr1_tr2 = -9

        # altenative time diff in case origin time is not listed
        starttime_diff_tr1_tr2 = tr1.stats.starttime - tr2.stats.starttime
        log.debug("# origin_time_diff_tr1_tr2 = {0}".format(origin_time_diff_tr1_tr2))
        log.debug("# starttime_diff_tr1_tr2 = {0}".format(starttime_diff_tr1_tr2))

        origin_time_diff_tr1_tr2_out = "{:.3e}".format(origin_time_diff_tr1_tr2)
        starttime_diff_tr1_tr2_out = "{:.3e}".format(starttime_diff_tr1_tr2)

        
        fs1 = tr1.stats.sampling_rate
        fs2 = tr2.stats.sampling_rate

        log.debug("# fs1 = {0} fs2 = {1}".format(fs1, fs2))

        
        evla1, evlo1, evdp1, mag1, evid1, stla1, stlo1 = sac_header_info(tr1)
        evla2, evlo2, evdp2, mag2, evid2, stla2, stlo2 = sac_header_info(tr2)
        
        fsOPT = 1
        diff_fs_abs = np.abs(fs1-fs2) 
        if diff_fs_eps < diff_fs_abs:
            log.warning("sampling rates do not match ! evid = {0} fs1 = {1} evid2 = {2} fs2 = {3}"
                        .format(evid1, fs1, evid2, fs2))
            fsOPT = 0
        
        

        log.debug("# evla1 = {0} evlo1 = {1} evdp1 = {2}".format(evla1, evlo1, evdp1))
        log.debug("# mag1 = = {0} evid1 = {1}".format(mag1, evid1))
        log.debug("# stla1 = {0} stlo1 = {1}".format(stla1, stlo1))

        log.debug("# evla2 = {0} evlo2 = {1} evdp2 = {2}".format(evla2, evlo2, evdp2))
        log.debug("# mag2 = = {0} evid2 = {1}".format(mag2, evid2))
        log.debug("# stla2 = {0} stlo2 = {1}".format(stla2, stlo2))

        mag_diff = mag1 - mag2
        mag_diff_out = "{:.2f}".format(mag_diff)


        geod = ppj.Geod(ellps='WGS84')
        az12, az21, ev_dist_m = geod.inv(evlo1, evla1, evlo2, evla2)
        ev_dist_km = ev_dist_m/1000.0
        
        log.debug("# ev_dist_km = {0}".format(ev_dist_km))
        
        # sta1 - ev dist 3D
        az12, az21, st1_ev_dist_m = geod.inv(evlo1, evla1, stlo1, stla1)
        st1_ev_dist_km = st1_ev_dist_m/1000.0
        # evdp is in km
        # stdv = 0
        st1_ev_3Ddist_km = np.sqrt( (st1_ev_dist_km*st1_ev_dist_km) + (evdp1*evdp1)   )

        # sta2 - ev dist 3D
        az12, az21, st2_ev_dist_m = geod.inv(evlo2, evla2, stlo2, stla2)
        st2_ev_dist_km = st2_ev_dist_m/1000.0
        # evdp is in km
        # stdv = 0
        st2_ev_3Ddist_km = np.sqrt( (st2_ev_dist_km*st2_ev_dist_km) + (evdp2*evdp2)   )

        ev_3Ddist_km_max = np.amax(np.array([st1_ev_3Ddist_km, st2_ev_3Ddist_km]))
        log.debug("ev_3Ddist_km_max = {0} st1_ev_3Ddist_km = {1} st2_ev_3Ddist_km = {2}".format(ev_3Ddist_km_max, st1_ev_3Ddist_km, st2_ev_3Ddist_km))

        if tw_xc_opt == "static":
            tw_xc_out = tw_xc_static
        if tw_xc_opt == "dynamic":
            tw_xc_out = (ev_3Ddist_km_max/(2.0*tw_xc_vs) ) + tw_xc_dynamic
        log.debug("tw_xc_opt = {0} tw_xc_out = {1}".format(tw_xc_opt, tw_xc_out))

        ev_3Ddist_km_max_out = "{:.3f}".format(ev_3Ddist_km_max)
        tw_xc_out_out = "{:.3f}".format(tw_xc_out)


        coh_start_time1 = tr1.stats.starttime + (tr1.stats.sac['t5'] - tr1.stats.sac['b'] - pre_tw_xc)
        coh_end_time1 = coh_start_time1 + tw_xc_out - tr1.stats.delta
        coh_start_time2 = tr2.stats.starttime + (tr2.stats.sac['t5'] - tr2.stats.sac['b'] - pre_tw_xc)
        coh_end_time2 = coh_start_time2+ tw_xc_out - tr2.stats.delta
        
        log.debug("coh_start_time1 = {0} coh_end_time1 = {1}".format(coh_start_time1, coh_end_time1))
        log.debug("coh_start_time2 = {0} coh_end_time2 = {1}".format(coh_start_time2, coh_end_time2))

        # zero padding
        tr1_trim = tr1.copy().trim(starttime=coh_start_time1, endtime=coh_end_time1, pad=True, fill_value=0)
        # zero padding
        tr2_trim = tr2.copy().trim(starttime=coh_start_time2, endtime=coh_end_time2, pad=True, fill_value=0)


        log.debug("# tr1_trim = {0}".format(tr1_trim))
        log.debug("# tr1_trim.stats.npts = {0}".format(tr1_trim.stats.npts))
        log.debug("# tr2_trim = {0}".format(tr2_trim))

        ev_distOPT = 1
        if ev_dist_km > max_ev_dist_km:
            log.debug("# no coh compute ev_dist_km = {0} max_ev_dist_km = {1}".format(ev_dist_km, max_ev_dist_km)) 
            ev_distOPT = 0
            
            
        log.debug("# fsOPT = {0}".format(fsOPT))
        log.debug("# ev_distOPT = {0}".format(ev_distOPT))
                
        if fsOPT and ev_distOPT:


            cc_spearmanr_outINT, p_spearmanr, cc_pearsonr_outINT, p_pearsonr = corr_norm_stat(tr1_trim.data, tr2_trim.data)


            log.debug("# cc_pearsonr_outINT (used most time) = {0}".format(cc_pearsonr_outINT))
            log.debug("# p_pearsonr = {0}".format(p_pearsonr))

            log.debug("# cc_spearmanr_outINT = {0}".format(cc_spearmanr_outINT))
            log.debug("# p_spearmanr= {0}".format(p_spearmanr))


            f_sp, Cyx_sp = signal.coherence(tr1_trim.data, tr2_trim.data, fs1, nperseg=nperseg, noverlap=noverlap)

            coh_mean, coh_mean_outINT = coh_stat(f_sp, Cyx_sp, fl, fh)
            log.debug("# fl = {0} fh = {1}".format(fl,fh))
            log.debug("# coh_mean = {0}".format(coh_mean))
            log.debug("# coh_mean_outINT = {0}".format(coh_mean_outINT))

            coh_mean1, coh_mean1_outINT = coh_stat(f_sp, Cyx_sp, fl1, fh1)
            log.debug("# fl1 = {0} fh1 = {1}".format(fl1,fh1))
            log.debug("# coh_mean1 = {0}".format(coh_mean1))
            log.debug("# coh_mean1_outINT = {0}".format(coh_mean1_outINT))

            coh_mean2, coh_mean2_outINT = coh_stat(f_sp, Cyx_sp, fl2, fh2)
            log.debug("# fl2 = {0} fh2 = {1}".format(fl2,fh2))
            log.debug("# coh_mean2 = {0}".format(coh_mean2))
            log.debug("# coh_mean2_outINT = {0}".format(coh_mean2_outINT))


            coh_mean3, coh_mean3_outINT = coh_stat(f_sp, Cyx_sp, fl3, fh3)
            log.debug("# fl3 = {0} fh3 = {1}".format(fl3,fh3))
            log.debug("# coh_mean3 = {0}".format(coh_mean3))
            log.debug("# coh_mean3_outINT = {0}".format(coh_mean3_outINT))


            coh_mean4, coh_mean4_outINT = coh_stat(f_sp, Cyx_sp, fl4, fh4)
            log.debug("# fl4 = {0} fh4 = {1}".format(fl4,fh4))
            log.debug("# coh_mean4 = {0}".format(coh_mean4))
            log.debug("# coh_mean4_outINT = {0}".format(coh_mean4_outINT))


            pair_id = mseed_id1+"_"+mseed_id2
            #dict_coh = {pair_id:{'mseed_id1':mseed_id1, 'mseed_id2':mseed_id2, 'cc_pearsonr_outINT':cc_pearsonr_outINT, 'coh_mean_outINT':coh_mean_outINT, 
            #                         'evid1':evid1, 'evid2':evid2, 'coh_mean1_outINT':coh_mean1_outINT, 'coh_mean2_outINT':coh_mean2_outINT, 'coh_mean3_outINT':coh_mean3_outINT, 'coh_mean4_outINT':coh_mean4_outINT,
            #                          'mag1':mag1, 'mag2':mag2   } }
            #dict_coh = {'mseed_id1':mseed_id1, 'mseed_id2':mseed_id2, 'cc_pearsonr_outINT':cc_pearsonr_outINT, 'coh_mean_outINT':coh_mean_outINT, 
            #                        'evid1':evid1, 'evid2':evid2, 'coh_mean1_outINT':coh_mean1_outINT, 'coh_mean2_outINT':coh_mean2_outINT, 'coh_mean3_outINT':coh_mean3_outINT, 'coh_mean4_outINT':coh_mean4_outINT,
            #                        'mag1':mag1, 'mag2':mag2   } 
            dict_coh = {'mseed_id1':mseed_id1, 'mseed_id2':mseed_id2, 'cc_pearsonr_outINT':cc_pearsonr_outINT, 'coh_mean_outINT':coh_mean_outINT, 
                                    'evid1':evid1, 'evid2':evid2, 'coh_mean1_outINT':coh_mean1_outINT, 'coh_mean2_outINT':coh_mean2_outINT, 'coh_mean3_outINT':coh_mean3_outINT, 'coh_mean4_outINT':coh_mean4_outINT,
                                    'mag1':mag1, 'mag2':mag2, 'mag_diff':mag_diff_out,
                                    'tw_xc_out':tw_xc_out_out, 'ev_3Ddist_km_max':ev_3Ddist_km_max_out,
                                    'origin_time_diff_tr1_tr2': origin_time_diff_tr1_tr2_out, 'starttime_diff_tr2_tr2': starttime_diff_tr1_tr2_out}


            return dict_coh
    
    except Exception as e:
        log.error("cannot preprocess for {0} - {1} pair".format(sacfi_name1, sacfi_name2))
        log.error(e)

def start_parallel_jobs_inner_loop(tasks, do_work, njobs=0):
    if njobs == 0:
        results = [do_work(task) for task in tasks]
    else:
        pool = multiprocessing.Pool(njobs)
        results = pool.map(do_work, tasks)
        pool.close()
        pool.join()
        
    return results


def coh_xc(trace_id,
           data_top_dir, traces_dict, output_csv, 
           pre_tw_save, post_tw_save,
           fl_xc, fh_xc, fl1, fh1, fl2, fh2, 
           fl3, fh3, fl4, fh4, 
           pre_tw_xc, tw_xc_static, max_ev_dist_km,
           plotOPT,
           nperseg, noverlap,
           diff_fs_eps, search_opt,
           tw_xc_dynamic, tw_xc_vs, tw_xc_opt,
           njobs, task_limit):
        
        #log.info("fl_xc = {0}".format(fl_xc))
        #log.info("fh_xc = {0}".format(fh_xc))
        #log.info("data_top_dir = {0}".format(data_top_dir))

        
        log.info("trace_id = {0}".format(trace_id))

        log.info("data_top_dir = {0}".format(data_top_dir))

        log.info("traces_dict = {0}".format(traces_dict))
        log.info("output_csv = {0}".format(output_csv))

        log.info("fl_xc = {0} fl_xc = {1}".format(fl_xc, fh_xc))
        log.info("fl1 = {0} fl1 = {1}".format(fl1, fh1))
        log.info("fl2 = {0} fl2 = {1}".format(fl2, fh2))
        log.info("fl3 = {0} fl3 = {1}".format(fl3, fh3))
        log.info("fl4 = {0} fl4 = {1}".format(fl4, fh4))

        log.info("pre_tw_xc = {0}".format(pre_tw_xc))
        log.info("tw_xc_static = {0}".format(tw_xc_static))
        log.info("max_ev_dist_km = {0}".format(max_ev_dist_km))

        log.info("plotOPT = {0}".format(plotOPT))

        log.info("nperseg = {0}".format(nperseg))
        log.info("noverlap = {0}".format(noverlap))

        log.info("diff_fs_eps = {0}".format(diff_fs_eps))

        log.info("search_opt = {0}".format(search_opt))

        log.info("tw_xc_dynamic = {0}".format(tw_xc_dynamic))
        log.info("tw_xc_vs = {0}".format(tw_xc_vs))
        log.info("tw_xc_opt = {0}".format(tw_xc_opt))

        log.info("njobs = {0}".format(njobs))
        log.info("task_limit = {0}".format(task_limit))

     

        if os.path.isfile(output_csv):
            os.remove(output_csv)

        #BK_BKS_00_HHZ_1_40_25_50_data.all/
        trace_idOUT = trace_id.replace('.', '_')

        data_directory = data_top_dir+"/cs_xc."+trace_id+".all/"+trace_idOUT+"_"+str(fl_xc)+"_"+str(fh_xc)+"_"+str(pre_tw_save)+"_"+str(post_tw_save)+"_data.all"

        log.info("data_directory = {0}".format(data_directory))
        p = Path(data_directory)
        sac_fis = list(p.glob("*.sac"))

        # additional traces
        #traces_dict[trace_id]
        trace_ids = traces_dict.get(trace_id)
        log.info("trace_ids = {0}".format(trace_ids))
        log.info("type(trace_ids) = {0}".format(type(trace_ids)))

        if trace_ids is not None:
            for i, trace_id_additional in trace_ids.items():
                log.info("i, trace_id_additional = {0} {1}".format(i, trace_id_additional))

                trace_id_additionalOUT = trace_id_additional.replace('.', '_')

                data_directory_additional = data_top_dir+"/cs_xc."+trace_id_additional+".all/"+trace_id_additionalOUT+"_"+str(fl_xc)+"_"+str(fh_xc)+"_"+str(pre_tw_save)+"_"+str(post_tw_save)+"_data.all"

                log.info("data_directory_additional = {0}".format(data_directory_additional))
                p_additional = Path(data_directory_additional)
                sac_fis_additional = list(p_additional.glob("*.sac"))
                log.info("total_sac_fis_additional = {0}".format(len(sac_fis_additional)))

                sac_fis = sac_fis + sac_fis_additional

        log.info("total_sac_fis = {0}".format(len(sac_fis)))

        sac_fis_resolve_name_full_path = list(map(lambda x: str((x.resolve()).parents[0]) +"/"+ (x.resolve()).name  , sac_fis))

        if len(sac_fis_resolve_name_full_path) > 0:
            log.info("sac_fis_resolve_name_full_path[0] = {0}".format(sac_fis_resolve_name_full_path[0]))

        tasks = get_tasks(sac_fis_resolve_name_full_path, search_opt)
        if len(tasks) > 0:
            log.info("task[0] = {0}".format(tasks[0]))

        log.info("total_correlation = {0}".format(len(tasks)))

        #sys.exit()

        do_work = partial(cal_coh,
                          fl_xc, fh_xc, fl1, fh1, 
                          fl2, fh2, fl3, fh3, fl4, fh4, 
                          pre_tw_xc, tw_xc_static, max_ev_dist_km,
                          plotOPT,
                          nperseg, noverlap,
                          diff_fs_eps, 
                          tw_xc_dynamic, tw_xc_vs, tw_xc_opt)
        

        #task_limit = 10
        if len(tasks) > 0:
            write_times = 1 
            try:
                for i in range(0, len(tasks), task_limit):
                    log.info("slice from i {0} to i+task_limit {1} ".format(i, i+task_limit))
                    task_subset = (tasks[i: i+task_limit])

                    results = start_parallel_jobs_inner_loop(task_subset, do_work, njobs)
    
                    #results_filtered = [x for x in results if x is not None]
                    results_filtered = [dict(x) for x in results if x is not None]
                    log.info("len(results_filtered)= {0}".format(len(results_filtered)))  
                    if len(results_filtered) > 0:    
                        with open(output_csv,'a',encoding='utf-8',newline='') as csvfile:
                            writer = csv.DictWriter(csvfile, fieldnames = results_filtered[0].keys())
                            if write_times == 1:
                                writer.writeheader()
                            write_times = write_times + 1
                            writer.writerows(results_filtered)
            except Exception as e:
                log.error("cannot coh for all data")
                log.error(e)
    

            
        #return results_filtered

#%%
def main():

    parser = argparse.ArgumentParser(description='CRE_coh')   
    parser.add_argument('-t', '--trace_id', help='trace id', required=True)   
    parser.add_argument('-p', '--parameter_file', help='parameter file', default="CRE_coh.jsonc")   
    parser.add_argument('-tl', '--traces_list', help='traces list', default="CRE_traces_list.json")   
    parser.add_argument('--debug', action='store_true')

    args = parser.parse_args()    
    
    logging_level = "INFO"
    if args.debug:
        logging_level = "DEBUG"
  
    logging.basicConfig(level=logging_level, format="%(asctime)s\t%(name)s\t%(levelname)s\t%(message)s")
    log.info(log)
    log.info("args.trace_id = {0}".format(args.trace_id))


    #inp_fi = "CRE_coh.jsonc"
    inp_fi = args.parameter_file
    log.info("inp_fi = {0}".format(inp_fi))

    inp_fd = open(inp_fi, mode='r')
    
    #inp_para = json.load(inp_fd)
    inp_para = load_jsonc(inp_fi)
    inp_fd.close()
    log.info('inp_para = {0}'.format(inp_para))
    
    #traces_list_fi = "CRE_traces_list.json"
    traces_list_fi = args.traces_list
    log.info("traces_list_fi = {0}".format(traces_list_fi))

    traces_dict = {}
    with open(traces_list_fi, mode="r") as f:
        #traces_dict = json.load(f)
        traces_dict = json.load(f)

    #trace_list_fi = "CRE_traces_list.json"
    #trace_list_fd = open(trace_list_fi, mode='r')    
    #traces_dict = load_jsonc(trace_list_fd)
    
    log.info("traces_dict = {0}".format(traces_dict))
    #log.info("type(traces_dict) = {0}".format(type(traces_dict)))
    #sys.exit()

    coh_xc(args.trace_id,
           inp_para['data_top_dir'], traces_dict, inp_para['output_csv'],
           inp_para['pre_tw_save'], inp_para['post_tw_save'],
           inp_para['fl_xc'], inp_para['fh_xc'], inp_para['fl1'], inp_para['fh1'], 
           inp_para['fl2'], inp_para['fh2'], inp_para['fl3'], inp_para['fh3'], 
           inp_para['fl4'], inp_para['fh4'], 
           inp_para['pre_tw_xc'], inp_para['tw_xc_static'], inp_para['max_ev_dist_km'],
           inp_para['plotOPT'],
           inp_para['nperseg'], inp_para['noverlap'],
           inp_para['diff_fs_eps'], inp_para['search_opt'],
           inp_para['tw_xc_dynamic'], inp_para['tw_xc_vs'], inp_para['tw_xc_opt'],
           inp_para['njobs'], inp_para['task_limit'])


if __name__ == "__main__":
    main()

 
# %%
